from collections import defaultdict
import logging
import subprocess
import sys
from typing import Any, DefaultDict, Dict, List, Optional, Tuple, Union
import click
from functools import wraps
import typer
from typer.core import TyperGroup, TyperCommand, MarkupMode
from rich.console import Console
from rich.table import Table
from rich.text import Text

from safety.auth.constants import CLI_AUTH, MSG_NON_AUTHENTICATED
from safety.auth.models import Auth
from safety.constants import MSG_NO_AUTHD_CICD_PROD_STG, MSG_NO_AUTHD_CICD_PROD_STG_ORG, MSG_NO_AUTHD_DEV_STG, MSG_NO_AUTHD_DEV_STG_ORG_PROMPT, MSG_NO_AUTHD_DEV_STG_PROMPT, MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL, MSG_NO_VERIFIED_EMAIL_TPL
from safety.scan.constants import CONSOLE_HELP_THEME

from safety.scan.models import ScanOutput

from .util import output_exception
from .errors import SafetyError, SafetyException

LOG = logging.getLogger(__name__)

def custom_print_options_panel(name: str, params: List[Any], ctx: Any, console: Console) -> None:
    """
    Print a panel with options.

    Args:
        name (str): The title of the panel.
        params (List[Any]): The list of options to print.
        ctx (Any): The context object.
        markup_mode (str): The markup mode.
        console (Console): The console to print to.
    """
    table = Table(title=name, show_lines=True)
    for param in params:
        table.add_row(str(param.opts), param.help or "")
    console.print(table)

def custom_print_commands_panel(name: str, commands: List[Any], console: Console) -> None:
    """
    Print a panel with commands.

    Args:
        name (str): The title of the panel.
        commands (List[Any]): The list of commands to print.
        console (Console): The console to print to.
    """
    table = Table(title=name, show_lines=True)
    for command in commands:
        table.add_row(command.name, command.help or "")
    console.print(table)

def custom_make_rich_text(text: str) -> Text:
    """
    Create rich text.

    Args:
        text (str): The text to format.

    Returns:
        Text: The formatted rich text.
    """
    return Text(text)

def custom_get_help_text(obj: Any) -> Text:
    """
    Get the help text for an object.

    Args:
        obj (Any): The object to get help text for.

    Returns:
        Text: The formatted help text.
    """
    return Text(obj.help)

def custom_make_command_help(help_text: str) -> Text:
    """
    Create rich text for command help.

    Args:
        help_text (str): The help text to format.
        markup_mode (str): The markup mode.

    Returns:
        Text: The formatted rich text.
    """
    return Text(help_text)

def get_command_for(name: str, typer_instance: typer.Typer) -> click.Command:
    """
    Retrieve a command by name from a Typer instance.

    Args:
        name (str): The name of the command.
        typer_instance (typer.Typer): The Typer instance.

    Returns:
        click.Command: The found command.
    """
    single_command = next(
        (command
         for command in typer_instance.registered_commands
         if command.name == name), None)

    if not single_command:
        raise ValueError("Unable to find the command name.")

    single_command.context_settings = typer_instance.info.context_settings
    click_command = typer.main.get_command_from_info(
        single_command,
        pretty_exceptions_short=typer_instance.pretty_exceptions_short,
        rich_markup_mode=typer_instance.rich_markup_mode,
    )
    if typer_instance._add_completion:
        click_install_param, click_show_param = \
            typer.main.get_install_completion_arguments()
        click_command.params.append(click_install_param)
        click_command.params.append(click_show_param)
    return click_command


def pass_safety_cli_obj(func):
    """
    Decorator to ensure the SafetyCLI object exists for a command.
    """
    @wraps(func)
    def inner(ctx, *args, **kwargs):

        if not ctx.obj:
            from .models import SafetyCLI
            ctx.obj = SafetyCLI()

        return func(ctx, *args, **kwargs)

    return inner


def pretty_format_help(obj: Union[click.Command, click.Group],
                       ctx: click.Context, markup_mode: MarkupMode) -> None:
    """
    Format and print help text in a pretty format.

    Args:
        obj (Union[click.Command, click.Group]): The Click command or group.
        ctx (click.Context): The Click context.
        markup_mode (MarkupMode): The markup mode.
    """
    from typer.rich_utils import highlighter, STYLE_USAGE_COMMAND, \
            ARGUMENTS_PANEL_TITLE, OPTIONS_PANEL_TITLE, \
                COMMANDS_PANEL_TITLE
    from rich.align import Align
    from rich.padding import Padding
    from rich.console import Console
    from rich.theme import Theme

    console = Console()

    with console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console
        # Print command / group help if we have some
        if obj.help:
            console.print()

            # Print with some padding
            console.print(
                Padding(
                    Align(custom_get_help_text(obj=obj), pad=False),
                    (0, 1, 0, 1)
                )
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            panel_to_commands: DefaultDict[str, List[click.Command]] = defaultdict(list)
            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    panel_name = (
                        getattr(command, "rich_help_panel", None)
                        or COMMANDS_PANEL_TITLE
                    )
                    panel_to_commands[panel_name].append(command)

            # Print each command group panel
            default_commands = panel_to_commands.get(COMMANDS_PANEL_TITLE, [])
            custom_print_commands_panel(
                name=COMMANDS_PANEL_TITLE,
                commands=default_commands,
                console=console,
            )
            for panel_name, commands in panel_to_commands.items():
                if panel_name == COMMANDS_PANEL_TITLE:
                    # Already printed above
                    continue
                custom_print_commands_panel(
                    name=panel_name,
                    commands=commands,
                    console=console,
                )

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)

        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                console=console,
            )

        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                console=console,
            )

        if ctx.parent:
            params = []
            for param in ctx.parent.command.params:
                if isinstance(param, click.Option):
                    params.append(param)

            custom_print_options_panel(
                name="Global-Options",
                params=params,
                ctx=ctx.parent,
                console=console,
            )

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = custom_make_rich_text(text=epilogue)
            console.print(Padding(Align(epilogue_text, pad=False), 1))



def print_main_command_panels(*,
    name: str,
    commands: List[click.Command],
    markup_mode: MarkupMode,
    console) -> None:
    """
    Print the main command panels.

    Args:
        name (str): The name of the panel.
        commands (List[click.Command]): List of commands to display.
        markup_mode (MarkupMode): The markup mode.
        console: The Rich console.
    """
    from rich import box
    from rich.table import Table
    from rich.text import Text
    from rich.panel import Panel
    from typer.rich_utils import STYLE_COMMANDS_TABLE_SHOW_LINES, STYLE_COMMANDS_TABLE_LEADING, \
        STYLE_COMMANDS_TABLE_BOX, STYLE_COMMANDS_TABLE_BORDER_STYLE, STYLE_COMMANDS_TABLE_ROW_STYLES, \
            STYLE_COMMANDS_TABLE_PAD_EDGE, STYLE_COMMANDS_TABLE_PADDING, STYLE_COMMANDS_PANEL_BORDER, \
                ALIGN_COMMANDS_PANEL

    t_styles: Dict[str, Any] = {
        "show_lines": STYLE_COMMANDS_TABLE_SHOW_LINES,
        "leading": STYLE_COMMANDS_TABLE_LEADING,
        "box": STYLE_COMMANDS_TABLE_BOX,
        "border_style": STYLE_COMMANDS_TABLE_BORDER_STYLE,
        "row_styles": STYLE_COMMANDS_TABLE_ROW_STYLES,
        "pad_edge": STYLE_COMMANDS_TABLE_PAD_EDGE,
        "padding": STYLE_COMMANDS_TABLE_PADDING,
    }
    box_style = getattr(box, t_styles.pop("box"), None)

    commands_table = Table(
        highlight=False,
        show_header=False,
        expand=True,
        box=box_style,
        **t_styles,
    )

    console_width = 80
    column_width = 25

    if console.size and console.size[0] > 80:
        console_width = console.size[0]

    commands_table.add_column(style="bold cyan", no_wrap=True, width=column_width, max_width=column_width)
    commands_table.add_column(width=console_width - column_width)

    rows = []

    for command in commands:
        helptext = command.short_help or command.help or ""
        command_name = command.name or ""
        command_name_text = Text(command_name)
        rows.append(
            [
                command_name_text,
                custom_make_command_help(
                    help_text=helptext,
                ),
            ]
        )
        rows.append([])
    for row in rows:
        commands_table.add_row(*row)
    if commands_table.row_count:
        console.print(
            Panel(
                commands_table,
                border_style=STYLE_COMMANDS_PANEL_BORDER,
                title=name,
                title_align=ALIGN_COMMANDS_PANEL,
            )
        )

# The help output for the main safety root command: `safety --help`
def format_main_help(obj: Union[click.Command, click.Group],
                       ctx: click.Context, markup_mode: MarkupMode) -> None:
    """
    Format the main help output for the safety root command.

    Args:
        obj (Union[click.Command, click.Group]): The Click command or group.
        ctx (click.Context): The Click context.
        markup_mode (MarkupMode): The markup mode.
    """
    from typer.rich_utils import highlighter, STYLE_USAGE_COMMAND, \
        ARGUMENTS_PANEL_TITLE, OPTIONS_PANEL_TITLE, \
            COMMANDS_PANEL_TITLE

    from rich.align import Align
    from rich.padding import Padding
    from rich.console import Console
    from rich.theme import Theme

    typer_console = Console()

    with typer_console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console

        # Print command / group help if we have some
        if obj.help:
            console.print()
            # Print with some padding
            console.print(
                Padding(
                    Align(custom_get_help_text(obj=obj),
                        pad=False,
                    ),
                    (0, 1, 0, 1),
                )
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            UTILITY_COMMANDS_PANEL_TITLE = "Commands cont."

            panel_to_commands: DefaultDict[str, List[click.Command]] = defaultdict(list)
            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    panel_name = (
                        UTILITY_COMMANDS_PANEL_TITLE if command.utility_command else COMMANDS_PANEL_TITLE
                    )
                    panel_to_commands[panel_name].append(command)

            # Print each command group panel
            default_commands = panel_to_commands.get(COMMANDS_PANEL_TITLE, [])
            print_main_command_panels(
                name=COMMANDS_PANEL_TITLE,
                commands=default_commands,
                markup_mode=markup_mode,
                console=console,
            )
            for panel_name, commands in panel_to_commands.items():
                if panel_name == COMMANDS_PANEL_TITLE:
                    # Already printed above
                    continue
                print_main_command_panels(
                    name=panel_name,
                    commands=commands,
                    markup_mode=markup_mode,
                    console=console,
                )

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)
        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                console=console,
            )
        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                console=console,
            )

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = custom_make_rich_text(text=epilogue)
            console.print(Padding(Align(epilogue_text, pad=False), 1))


def process_auth_status_not_ready(console, auth: Auth, ctx: typer.Context) -> None:
    """
    Handle the process when the authentication status is not ready.

    Args:
        console: The Rich console.
        auth (Auth): The Auth object.
        ctx (typer.Context): The Typer context.
    """
    from safety_schemas.models import Stage
    from rich.prompt import Confirm, Prompt

    if not auth.client or not auth.client.is_using_auth_credentials():

        if auth.stage is Stage.development:
            console.print()
            if auth.org:
                confirmed = Confirm.ask(MSG_NO_AUTHD_DEV_STG_ORG_PROMPT, choices=["Y", "N", "y", "n"],
                                        show_choices=False, show_default=False,
                                        default=True, console=console)

                if not confirmed:
                    sys.exit(0)

                from safety.auth.cli import auth_app
                login_command = get_command_for(name='login',
                                        typer_instance=auth_app)
                ctx.invoke(login_command)
            else:
                console.print(MSG_NO_AUTHD_DEV_STG)
                console.print()
                choices = ["L", "R", "l", "r"]
                next_command = Prompt.ask(MSG_NO_AUTHD_DEV_STG_PROMPT, default=None,
                                          choices=choices, show_choices=False,
                                          console=console)

                from safety.auth.cli import auth_app
                login_command = get_command_for(name='login',
                                        typer_instance=auth_app)
                register_command = get_command_for(name='register',
                                        typer_instance=auth_app)
                if next_command is None or next_command.lower() not in choices:
                    sys.exit(0)

                console.print()
                if next_command.lower() == "r":
                    ctx.invoke(register_command)
                else:
                    ctx.invoke(login_command)

                if not ctx.obj.auth.email_verified:
                    sys.exit(1)
        else:
            if not auth.org:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG_ORG.format(LOGIN_URL=CLI_AUTH))

            else:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG)
                console.print(
                    MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL.format(
                        LOGIN_URL=CLI_AUTH,
                        SIGNUP_URL=f"{CLI_AUTH}/?sign_up=True"))
            sys.exit(1)

    elif not auth.email_verified:
        console.print()
        console.print(MSG_NO_VERIFIED_EMAIL_TPL.format(email=auth.email if auth.email else "Missing email"))
        sys.exit(1)
    else:
        console.print(MSG_NON_AUTHENTICATED)
        sys.exit(1)

class UtilityCommandMixin:
    """
    Mixin to add utility command functionality.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the UtilityCommandMixin.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        self.utility_command = kwargs.pop('utility_command', False)
        super().__init__(*args, **kwargs)

class SafetyCLISubGroup(UtilityCommandMixin, TyperGroup):
    """
    Custom TyperGroup with additional functionality for Safety CLI.
    """

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format usage message.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"

        formatter.write_usage(command_path, " ".join(pieces))

    def command(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> click.Command:
        """
        Create a new command.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            click.Command: The created command.
        """
        super().command(*args, **kwargs)

class SafetyCLICommand(UtilityCommandMixin, TyperCommand):
    """
    Custom TyperCommand with additional functionality for Safety CLI.
    """
    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format usage message.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"

        formatter.write_usage(command_path, " ".join(pieces))


class SafetyCLIUtilityCommand(TyperCommand):
    """
    Custom TyperCommand designated as a utility command.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the SafetyCLIUtilityCommand.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        self.utility_command = True
        super().__init__(*args, **kwargs)

class SafetyCLILegacyGroup(UtilityCommandMixin, click.Group):
    """
    Custom Click Group to handle legacy command-line arguments.
    """

    def parse_legacy_args(self, args: List[str]) -> Tuple[Optional[Dict[str, str]], Optional[str]]:
        """
        Parse legacy command-line arguments for proxy settings and keys.

        Args:
            args (List[str]): List of command-line arguments.

        Returns:
            Tuple[Optional[Dict[str, str]], Optional[str]]: Parsed proxy options and key.
        """
        options = {
            'proxy_protocol': 'https',
            'proxy_port': 80,
            'proxy_host': None
        }
        key = None

        for i, arg in enumerate(args):
            if arg in ['--proxy-protocol', '-pr'] and i + 1 < len(args):
                options['proxy_protocol'] = args[i + 1]
            elif arg in ['--proxy-port', '-pp'] and i + 1 < len(args):
                options['proxy_port'] = int(args[i + 1])
            elif arg in ['--proxy-host', '-ph'] and i + 1 < len(args):
                options['proxy_host'] = args[i + 1]
            elif arg in ['--key'] and i + 1 < len(args):
                key = args[i + 1]

        proxy = options if options['proxy_host'] else None
        return proxy, key

    def invoke(self, ctx: click.Context) -> None:
        """
        Invoke the command, handling legacy arguments.

        Args:
            ctx (click.Context): Click context.
        """
        args = ctx.args

        # Workaround for legacy check options, that now are global options
        subcommand_args = set(args)
        PROXY_HOST_OPTIONS = set(["--proxy-host", "-ph"])
        if "check" in ctx.protected_args or "license" in ctx.protected_args and (bool(PROXY_HOST_OPTIONS.intersection(subcommand_args) or "--key" in subcommand_args)) :
            proxy_options, key = self.parse_legacy_args(args)
            if proxy_options:
                ctx.params.update(proxy_options)

            if key:
                ctx.params.update({"key": key})

        # Now, invoke the original behavior
        super(SafetyCLILegacyGroup, self).invoke(ctx)


    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        # The main `safety --help`
        if self.name == "cli":
            format_main_help(self, ctx, markup_mode="rich")
        # All other help outputs
        else:
            pretty_format_help(self, ctx, markup_mode="rich")

class SafetyCLILegacyCommand(UtilityCommandMixin, click.Command):
    """
    Custom Click Command to handle legacy command-line arguments.
    """
    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode="rich")


def handle_cmd_exception(func):
    """
    Decorator to handle exceptions in command functions.

    Args:
        func: The command function to wrap.

    Returns:
        The wrapped function.
    """
    @wraps(func)
    def inner(ctx, output: Optional[ScanOutput], *args, **kwargs):
        if output:
            kwargs.update({"output": output})

            if output is ScanOutput.NONE:
                return func(ctx, *args, **kwargs)

        try:
            return func(ctx, *args, **kwargs)
        except click.ClickException as e:
            raise e
        except SafetyError as e:
            LOG.exception('Expected SafetyError happened: %s', e)
            output_exception(e, exit_code_output=True)
        except Exception as e:
            LOG.exception('Unexpected Exception happened: %s', e)
            exception = e if isinstance(e, SafetyException) else SafetyException(info=e)
            output_exception(exception, exit_code_output=True)

    return inner

def get_git_branch_name() -> Optional[str]:
    """
    Retrieves the current Git branch name.

    Returns:
        str: The current Git branch name, or None if it cannot be determined.
    """
    try:
        branch_name = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stderr=subprocess.DEVNULL,
            text=True
        ).strip()
        return branch_name if branch_name else None
    except Exception:
        return None




from unittest.mock import Mock, PropertyMock, patch, ANY
import click
from click.testing import CliRunner
import unittest

from safety.cli import cli
from safety.cli_util import get_command_for


class TestSafetyAuthCLI(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.runner = CliRunner(mix_stderr=False)

    @patch("safety.auth.cli.fail_if_authenticated")
    @patch("safety.auth.cli.get_authorization_data")
    @patch("safety.auth.cli.process_browser_callback")
    def test_auth_calls_login(self, process_browser_callback, 
                              get_authorization_data, fail_if_authenticated):
        auth_data = "https://safetycli.com", "initialState"
        get_authorization_data.return_value = auth_data
        process_browser_callback.return_value = {"email": "user@safetycli.com", "name": "Safety User"}
        result = self.runner.invoke(cli, ['auth'])

        fail_if_authenticated.assert_called_once()
        get_authorization_data.assert_called_once()
        process_browser_callback.assert_called_once_with(auth_data[0], 
                                                         initial_state=auth_data[1], 
                                                         ctx=ANY, headless=False)
        
        expected = [
            "",
            "Redirecting your browser to log in; once authenticated, return here to start using Safety",
            "",
            "You're authenticated",
            " Account: Safety User, user@safetycli.com (email verification required)",
            "",
            "To complete your account open the “verify your email” email sent to",
            "user@safetycli.com",
            "",
            "Can’t find the verification email? Login at",
            "`https://platform.safetycli.com/login/` to resend the verification email",
            ""
        ]

        for res_line, exp_line in zip(result.stdout.splitlines(), expected):
          self.assertIn(exp_line, res_line)



import logging
import os
import platform
import re
import sys
from collections import defaultdict
from datetime import datetime
from difflib import SequenceMatcher
from threading import Lock
from typing import List, Optional, Dict, Generator, Tuple, Union, Any

import click
from click import BadParameter
from dparse import parse, filetypes
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version
from packaging.specifiers import SpecifierSet
from requests import PreparedRequest
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError

from safety.constants import EXIT_CODE_FAILURE, EXIT_CODE_OK, HASH_REGEX_GROUPS, SYSTEM_CONFIG_DIR, USER_CONFIG_DIR
from safety.errors import InvalidProvidedReportError
from safety.models import Package, RequirementFile, is_pinned_requirement, SafetyRequirement
from safety_schemas.models import TelemetryModel

LOG = logging.getLogger(__name__)


def is_a_remote_mirror(mirror: str) -> bool:
    """
    Check if a mirror URL is remote.

    Args:
        mirror (str): The mirror URL.

    Returns:
        bool: True if the mirror URL is remote, False otherwise.
    """
    return mirror.startswith("http://") or mirror.startswith("https://")


def is_supported_by_parser(path: str) -> bool:
    """
    Check if the file path is supported by the parser.

    Args:
        path (str): The file path.

    Returns:
        bool: True if the file path is supported, False otherwise.
    """
    supported_types = (".txt", ".in", ".yml", ".ini", "Pipfile",
                       "Pipfile.lock", "setup.cfg", "poetry.lock")
    return path.endswith(supported_types)


def parse_requirement(dep: Any, found: str) -> SafetyRequirement:
    """
    Parse a requirement.

    Args:
        dep (Any): The dependency.
        found (str): The location where the dependency was found.

    Returns:
        SafetyRequirement: The parsed requirement.
    """
    req = SafetyRequirement(dep)
    req.found = found

    if req.specifier == SpecifierSet(''):
        req.specifier = SpecifierSet('>=0')

    return req


def find_version(requirements: List[SafetyRequirement]) -> Optional[str]:
    """
    Find the version of a requirement.

    Args:
        requirements (List[SafetyRequirement]): The list of requirements.

    Returns:
        Optional[str]: The version if found, None otherwise.
    """
    ver = None

    if len(requirements) != 1:
        return ver

    specs = requirements[0].specifier

    if is_pinned_requirement(specs):
        ver = next(iter(requirements[0].specifier)).version

    return ver


def read_requirements(fh: Any, resolve: bool = True) -> Generator[Package, None, None]:
    """
    Reads requirements from a file-like object and (optionally) from referenced files.

    Args:
        fh (Any): The file-like object to read from.
        resolve (bool): Resolves referenced files.

    Returns:
        Generator: Yields Package objects.
    """
    is_temp_file = not hasattr(fh, 'name')
    path = None
    found = 'temp_file'
    file_type = filetypes.requirements_txt
    absolute_path: Optional[str] = None

    if not is_temp_file and is_supported_by_parser(fh.name):
        LOG.debug('not temp and a compatible file')
        path = fh.name
        absolute_path = os.path.abspath(path)
        SafetyContext().scanned_full_path.append(absolute_path)
        found = path
        file_type = None

    LOG.debug(f'Path: {path}')
    LOG.debug(f'File Type: {file_type}')
    LOG.debug('Trying to parse file using dparse...')
    content = fh.read()
    LOG.debug(f'Content: {content}')
    dependency_file = parse(content, path=path, resolve=resolve,
                            file_type=file_type)
    LOG.debug(f'Dependency file: {dependency_file.serialize()}')
    LOG.debug(f'Parsed, dependencies: {[dep.serialize() for dep in dependency_file.resolved_dependencies]}')

    reqs_pkg = defaultdict(list)

    for req in dependency_file.resolved_dependencies:
        reqs_pkg[canonicalize_name(req.name)].append(req)

    for pkg, reqs in reqs_pkg.items():
        requirements = list(map(lambda req: parse_requirement(req, absolute_path), reqs))
        version = find_version(requirements)

        yield Package(name=pkg, version=version,
                      requirements=requirements,
                      found=found,
                      absolute_path=absolute_path,
                      insecure_versions=[],
                      secure_versions=[], latest_version=None,
                      latest_version_without_known_vulnerabilities=None,
                      more_info_url=None)


def get_proxy_dict(proxy_protocol: str, proxy_host: str, proxy_port: int) -> Optional[Dict[str, str]]:
    """
    Get the proxy dictionary for requests.

    Args:
        proxy_protocol (str): The proxy protocol.
        proxy_host (str): The proxy host.
        proxy_port (int): The proxy port.

    Returns:
        Optional[Dict[str, str]]: The proxy dictionary if all parameters are provided, None otherwise.
    """
    if proxy_protocol and proxy_host and proxy_port:
        # Safety only uses https request, so only https dict will be passed to requests
        return {'https': f"{proxy_protocol}://{proxy_host}:{str(proxy_port)}"}
    return None


def get_license_name_by_id(license_id: int, db: Dict[str, Any]) -> Optional[str]:
    """
    Get the license name by its ID.

    Args:
        license_id (int): The license ID.
        db (Dict[str, Any]): The database containing license information.

    Returns:
        Optional[str]: The license name if found, None otherwise.
    """
    licenses = db.get('licenses', [])
    for name, id in licenses.items():
        if id == license_id:
            return name
    return None


def get_flags_from_context() -> Dict[str, str]:
    """
    Get the flags from the current click context.

    Returns:
        Dict[str, str]: A dictionary of flags and their corresponding option names.
    """
    flags = {}
    context = click.get_current_context(silent=True)

    if context:
        for option in context.command.params:
            flags_per_opt = option.opts + option.secondary_opts
            for flag in flags_per_opt:
                flags[flag] = option.name

    return flags


def get_used_options() -> Dict[str, Dict[str, int]]:
    """
    Get the used options from the command-line arguments.

    Returns:
        Dict[str, Dict[str, int]]: A dictionary of used options and their counts.
    """
    flags = get_flags_from_context()
    used_options = {}

    for arg in sys.argv:
        cleaned_arg = arg if '=' not in arg else arg.split('=')[0]
        if cleaned_arg in flags:
            option_used = flags.get(cleaned_arg)

            if option_used in used_options:
                used_options[option_used][cleaned_arg] = used_options[option_used].get(cleaned_arg, 0) + 1
            else:
                used_options[option_used] = {cleaned_arg: 1}

    return used_options


def get_safety_version() -> str:
    """
    Get the version of Safety.

    Returns:
        str: The Safety version.
    """
    from safety import VERSION
    return VERSION


def get_primary_announcement(announcements: List[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    """
    Get the primary announcement from a list of announcements.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.

    Returns:
        Optional[Dict[str, Any]]: The primary announcement if found, None otherwise.
    """
    for announcement in announcements:
        if announcement.get('type', '').lower() == 'primary_announcement':
            try:
                from safety.output_utils import build_primary_announcement
                build_primary_announcement(announcement, columns=80)
            except Exception as e:
                LOG.debug(f'Failed to build primary announcement: {str(e)}')
                return None

            return announcement

    return None


def get_basic_announcements(announcements: List[Dict[str, Any]], include_local: bool = True) -> List[Dict[str, Any]]:
    """
    Get the basic announcements from a list of announcements.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.
        include_local (bool): Whether to include local announcements.

    Returns:
        List[Dict[str, Any]]: The list of basic announcements.
    """
    return [announcement for announcement in announcements if
            announcement.get('type', '').lower() != 'primary_announcement' and not announcement.get('local', False)
            or (announcement.get('local', False) and include_local)]


def filter_announcements(announcements: List[Dict[str, Any]], by_type: str = 'error') -> List[Dict[str, Any]]:
    """
    Filter announcements by type.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.
        by_type (str): The type of announcements to filter by.

    Returns:
        List[Dict[str, Any]]: The filtered announcements.
    """
    return [announcement for announcement in announcements if
            announcement.get('type', '').lower() == by_type]


def build_telemetry_data(telemetry: bool = True,
                         command: Optional[str] = None,
                         subcommand: Optional[str] = None) -> TelemetryModel:
    """Build telemetry data for the Safety context.

    Args:
        telemetry (bool): Whether telemetry is enabled.
        command (Optional[str]): The command.
        subcommand (Optional[str]): The subcommand.

    Returns:
        TelemetryModel: The telemetry data model.
    """
    context = SafetyContext()

    body = {
        'os_type': os.environ.get("SAFETY_OS_TYPE", None) or platform.system(),
        'os_release': os.environ.get("SAFETY_OS_RELEASE", None) or platform.release(),
        'os_description': os.environ.get("SAFETY_OS_DESCRIPTION", None) or platform.platform(),
        'python_version': platform.python_version(),
        'safety_command': command if command else context.command,
        'safety_options': get_used_options()
    } if telemetry else {}

    body['safety_version'] = get_safety_version()
    body['safety_source'] = os.environ.get("SAFETY_SOURCE", None) or context.safety_source

    if not 'safety_options' in body:
        body['safety_options'] = {}

    LOG.debug(f'Telemetry body built: {body}')

    return TelemetryModel(**body)


def build_git_data() -> Dict[str, Any]:
    """Build git data for the repository.

    Returns:
        Dict[str, str]: The git data.
    """
    import subprocess

    def git_command(commandline: List[str]) -> str:
        return subprocess.run(commandline, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.decode('utf-8').strip()

    try:
        is_git = git_command(["git", "rev-parse", "--is-inside-work-tree"])
    except Exception:
        is_git = False

    if is_git == "true":
        result = {
            "branch": "",
            "tag": "",
            "commit": "",
            "dirty": "",
            "origin": ""
        }

        try:
            result['branch'] = git_command(["git", "symbolic-ref", "--short", "-q", "HEAD"])
            result['tag'] = git_command(["git", "describe", "--tags", "--exact-match"])

            commit = git_command(["git", "describe", '--match=""', '--always', '--abbrev=40', '--dirty'])
            result['dirty'] = commit.endswith('-dirty')
            result['commit'] = commit.split("-dirty")[0]

            result['origin'] = git_command(["git", "remote", "get-url", "origin"])
        except Exception:
            pass

        return result
    else:
        return {
            "error": "not-git-repo"
        }


def output_exception(exception: Exception, exit_code_output: bool = True) -> None:
    """
    Output an exception message to the console and exit.

    Args:
        exception (Exception): The exception to output.
        exit_code_output (bool): Whether to output the exit code.

    Exits:
        Exits the program with the appropriate exit code.
    """
    click.secho(str(exception), fg="red", file=sys.stderr)

    if exit_code_output:
        exit_code = EXIT_CODE_FAILURE
        if hasattr(exception, 'get_exit_code'):
            exit_code = exception.get_exit_code()
    else:
        exit_code = EXIT_CODE_OK

    sys.exit(exit_code)

def build_remediation_info_url(base_url: str, version: Optional[str], spec: str,
                               target_version: Optional[str] = '') -> str:
    """
    Build the remediation info URL.

    Args:
        base_url (str): The base URL.
        version (Optional[str]): The current version.
        spec (str): The specification.
        target_version (Optional[str]): The target version.

    Returns:
        str: The remediation info URL.
    """

    params = {'from': version, 'to': target_version}

    # No pinned version
    if not version:
        params = {'spec': spec}

    req = PreparedRequest()
    req.prepare_url(base_url, params)

    return req.url

def get_processed_options(policy_file: Dict[str, Any], ignore: Dict[str, Any], ignore_severity_rules: Dict[str, Any],
    exit_code: bool, ignore_unpinned_requirements: Optional[bool] = None,
    project: Optional[str] = None) -> Tuple[Dict[str, Any], Dict[str, Any], bool, Optional[bool], Optional[str]]:
    """
    Get processed options from the policy file.

    Args:
        policy_file (Dict[str, Any]): The policy file.
        ignore (Dict[str, Any]): The ignore settings.
        ignore_severity_rules (Dict[str, Any]): The ignore severity rules.
        exit_code (bool): The exit code setting.
        ignore_unpinned_requirements (Optional[bool]): The ignore unpinned requirements setting.
        project (Optional[str]): The project setting.

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any], bool, Optional[bool], Optional[str]]: The processed options.
    """
    if policy_file:
        project_config = policy_file.get('project', {})
        security = policy_file.get('security', {})
        ctx = click.get_current_context()
        source = ctx.get_parameter_source("exit_code")

        if not project:
            project_id = project_config.get('id', None)
            if not project_id:
                project_id = None
            project = project_id

        if ctx.get_parameter_source("ignore_unpinned_requirements") == click.core.ParameterSource.DEFAULT:
            ignore_unpinned_requirements = security.get('ignore-unpinned-requirements', None)

        if not ignore:
            ignore = security.get('ignore-vulnerabilities', {})
        if source == click.core.ParameterSource.DEFAULT:
            exit_code = not security.get('continue-on-vulnerability-error', False)
        ignore_cvss_below = security.get('ignore-cvss-severity-below', 0.0)
        ignore_cvss_unknown = security.get('ignore-cvss-unknown-severity', False)
        ignore_severity_rules = {'ignore-cvss-severity-below': ignore_cvss_below,
                                 'ignore-cvss-unknown-severity': ignore_cvss_unknown}

    return ignore, ignore_severity_rules, exit_code, ignore_unpinned_requirements, project


def get_fix_options(policy_file: Dict[str, Any], auto_remediation_limit: int) -> int:
    """
    Get fix options from the policy file.

    Args:
        policy_file (Dict[str, Any]): The policy file.
        auto_remediation_limit (int): The auto remediation limit.

    Returns:
        int: The auto remediation limit.
    """
    auto_fix = []

    source = click.get_current_context().get_parameter_source("auto_remediation_limit")
    if source == click.core.ParameterSource.COMMANDLINE:
        return auto_remediation_limit

    if policy_file:
        fix = policy_file.get('security-updates', {})
        auto_fix = fix.get('auto-security-updates-limit', None)
        if not auto_fix:
            auto_fix = []

    return auto_fix


class MutuallyExclusiveOption(click.Option):
    """
    A click option that is mutually exclusive with other options.
    """

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop('mutually_exclusive', []))
        self.with_values = kwargs.pop('with_values', {})
        help = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(["{0} with values {1}".format(item, self.with_values.get(item)) if item in self.with_values else item for item in self.mutually_exclusive])
            kwargs['help'] = help + (
                ' NOTE: This argument is mutually exclusive with '
                ' arguments: [' + ex_str + '].'
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx: click.Context, opts: Dict[str, Any], args: List[str]) -> Tuple[Any, List[str]]:
        """
        Handle the parse result for mutually exclusive options.

        Args:
            ctx (click.Context): The click context.
            opts (Dict[str, Any]): The options dictionary.
            args (List[str]): The arguments list.

        Returns:
            Tuple[Any, List[str]]: The result and remaining arguments.
        """
        m_exclusive_used = self.mutually_exclusive.intersection(opts)
        option_used = m_exclusive_used and self.name in opts

        exclusive_value_used = False
        for used in m_exclusive_used:
            value_used = opts.get(used, None)
            if not isinstance(value_used, List):
                value_used = [value_used]
            if value_used and set(self.with_values.get(used, [])).intersection(value_used):
                exclusive_value_used = True

        if option_used and (not self.with_values or exclusive_value_used):
            options = ', '.join(self.opts)
            prohibited = ''.join(["\n * --{0} with {1}".format(item, self.with_values.get(
                        item)) if item in self.with_values else f"\n * {item}" for item in self.mutually_exclusive])
            raise click.UsageError(
                f"Illegal usage: `{options}` is mutually exclusive with: {prohibited}"
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


class DependentOption(click.Option):
    """
    A click option that depends on other options.
    """
    def __init__(self, *args, **kwargs):
        self.required_options = set(kwargs.pop('required_options', []))
        help = kwargs.get('help', '')
        if self.required_options:
            ex_str = ', '.join(self.required_options)
            kwargs['help'] = help + (
                f" Requires: [ {ex_str} ]"
            )
        super(DependentOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx: click.Context, opts: Dict[str, Any], args: List[str]) -> Tuple[Any, List[str]]:
        """
        Handle the parse result for dependent options.

        Args:
            ctx (click.Context): The click context.
            opts (Dict[str, Any]): The options dictionary.
            args (List[str]): The arguments list.

        Returns:
            Tuple[Any, List[str]]: The result and remaining arguments.
        """
        missing_required_arguments = None

        if self.name in opts:
            missing_required_arguments = self.required_options.difference(opts)

        if missing_required_arguments:
            raise click.UsageError(
                "Illegal usage: `{}` needs the "
                "arguments `{}`.".format(
                    self.name,
                    ', '.join(missing_required_arguments)
                )
            )

        return super(DependentOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


def transform_ignore(ctx: click.Context, param: click.Parameter, value: Tuple[str]) -> Dict[str, Dict[str, Optional[str]]]:
    """
    Transform ignore parameters into a dictionary.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (Tuple[str]): The parameter value.

    Returns:
        Dict[str, Dict[str, Optional[str]]]: The transformed ignore parameters.
    """
    ignored_default_dict = {'reason': '', 'expires': None}
    if isinstance(value, tuple) and any(value):
        # Following code is required to support the 2 ways of providing 'ignore'
        # --ignore=1234,567,789
        # or, the historical way (supported for backward compatibility)
        # -i 1234 -i 567
        combined_value = ','.join(value)
        ignore_ids = {vuln_id.strip() for vuln_id in combined_value.split(',')}
        return {ignore_id: dict(ignored_default_dict) for ignore_id in ignore_ids}

    return {}


def active_color_if_needed(ctx: click.Context, param: click.Parameter, value: str) -> str:
    """
    Activate color if needed based on the context and environment variables.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (str): The parameter value.

    Returns:
        str: The parameter value.
    """
    if value == 'screen':
        ctx.color = True

    color = os.environ.get("SAFETY_COLOR", None)

    if color is not None:
        color = color.lower()

        if color == '1' or color == 'true':
            ctx.color = True
        elif color == '0' or color == 'false':
            ctx.color = False

    return value


def json_alias(ctx: click.Context, param: click.Parameter, value: bool) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'json' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ['SAFETY_OUTPUT'] = 'json'
        return value

def html_alias(ctx: click.Context, param: click.Parameter, value: bool) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'html' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ['SAFETY_OUTPUT'] = 'html'
        return value


def bare_alias(ctx: click.Context, param: click.Parameter, value: bool) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'bare' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ['SAFETY_OUTPUT'] = 'bare'
        return value


def get_terminal_size() -> os.terminal_size:
    """
    Get the terminal size.

    Returns:
        os.terminal_size: The terminal size.
    """
    from shutil import get_terminal_size as t_size
    # get_terminal_size can report 0, 0 if run from pseudo-terminal prior Python 3.11 versions

    columns = t_size().columns or 80
    lines = t_size().lines or 24

    return os.terminal_size((columns, lines))


def clean_project_id(input_string: str) -> str:
    """
    Clean a project ID by removing non-alphanumeric characters and normalizing the string.

    Args:
        input_string (str): The input string.

    Returns:
        str: The cleaned project ID.
    """
    input_string = re.sub(r'[^a-zA-Z0-9]+', '-', input_string)
    input_string = input_string.strip('-')
    input_string = input_string.lower()

    return input_string


def validate_expiration_date(expiration_date: str) -> Optional[datetime]:
    """
    Validate an expiration date string.

    Args:
        expiration_date (str): The expiration date string.

    Returns:
        Optional[datetime]: The validated expiration date if valid, None otherwise.
    """
    d = None

    if expiration_date:
        try:
            d = datetime.strptime(expiration_date, '%Y-%m-%d')
        except ValueError as e:
            pass

        try:
            d = datetime.strptime(expiration_date, '%Y-%m-%d %H:%M:%S')
        except ValueError as e:
            pass

    return d


class SafetyPolicyFile(click.ParamType):
    """
       Custom Safety Policy file to hold validations.
    """

    name = "filename"
    envvar_list_splitter = os.path.pathsep

    def __init__(
        self,
        mode: str = "r",
        encoding: Optional[str] = None,
        errors: str = "strict",
        pure: bool = os.environ.get('SAFETY_PURE_YAML', 'false').lower() == 'true'
    ) -> None:
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.basic_msg = '\n' + click.style('Unable to load the Safety Policy file "{name}".', fg='red')
        self.pure = pure

    def to_info_dict(self) -> Dict[str, Any]:
        """
        Convert the object to an info dictionary.

        Returns:
            Dict[str, Any]: The info dictionary.
        """
        info_dict = super().to_info_dict()
        info_dict.update(mode=self.mode, encoding=self.encoding)
        return info_dict

    def fail_if_unrecognized_keys(self, used_keys: List[str], valid_keys: List[str], param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None, msg: str = '{hint}', context_hint: str = '') -> None:
        """
        Fail if unrecognized keys are found in the policy file.

        Args:
            used_keys (List[str]): The used keys.
            valid_keys (List[str]): The valid keys.
            param (Optional[click.Parameter]): The click parameter.
            ctx (Optional[click.Context]): The click context.
            msg (str): The error message template.
            context_hint (str): The context hint for the error message.

        Raises:
            click.UsageError: If unrecognized keys are found.
        """
        for keyword in used_keys:
            if keyword not in valid_keys:
                match = None
                max_ratio = 0.0
                if isinstance(keyword, str):
                    for option in valid_keys:
                        ratio = SequenceMatcher(None, keyword, option).ratio()
                        if ratio > max_ratio:
                            match = option
                            max_ratio = ratio

                maybe_msg = f' Maybe you meant: {match}' if max_ratio > 0.7 else \
                            f' Valid keywords in this level are: {", ".join(valid_keys)}'

                self.fail(msg.format(hint=f'{context_hint}"{keyword}" is not a valid keyword.{maybe_msg}'), param, ctx)

    def fail_if_wrong_bool_value(self, keyword: str, value: Any, msg: str = '{hint}') -> None:
        """
        Fail if a boolean value is invalid.

        Args:
            keyword (str): The keyword.
            value (Any): The value.
            msg (str): The error message template.

        Raises:
            click.UsageError: If the boolean value is invalid.
        """
        if value is not None and not isinstance(value, bool):
            self.fail(msg.format(hint=f"'{keyword}' value needs to be a boolean. "
                                      "You can use True, False, TRUE, FALSE, true or false"))

    def convert(self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]) -> Any:
        """
        Convert the parameter value to a Safety policy file.

        Args:
            value (Any): The parameter value.
            param (Optional[click.Parameter]): The click parameter.
            ctx (Optional[click.Context]): The click context.

        Returns:
            Any: The converted policy file.

        Raises:
            click.UsageError: If the policy file is invalid.
        """
        try:
            # Check if the value is already a file-like object
            if hasattr(value, "read") or hasattr(value, "write"):
                return value

            # Prepare the error message template
            msg = self.basic_msg.format(name=value) + '\n' + click.style('HINT:', fg='yellow') + ' {hint}'

            # Open the file stream
            f, _ = click.types.open_stream(
                value, self.mode, self.encoding, self.errors, atomic=False
            )
            filename = ''

            try:
                # Read the content of the file
                raw = f.read()
                yaml = YAML(typ='safe', pure=self.pure)
                safety_policy = yaml.load(raw)
                filename = f.name
                f.close()
            except Exception as e:
                # Handle YAML parsing errors
                show_parsed_hint = isinstance(e, MarkedYAMLError)
                hint = str(e)
                if show_parsed_hint:
                    hint = f'{str(e.problem).strip()} {str(e.context).strip()} {str(e.context_mark).strip()}'

                self.fail(msg.format(name=value, hint=hint), param, ctx)

            # Validate the structure of the safety policy
            if not safety_policy or not isinstance(safety_policy, dict) or not safety_policy.get('security', None):
                hint = "you are missing the security root tag"
                try:
                    version = safety_policy["version"]
                    if version:
                        hint = f"{filename} is a policy file version {version}. " \
                            "Legacy policy file parser only accepts versions minor than 3.0" \
                            "\nNote: `safety check` command accepts policy file versions <= 2.0. Versions >= 2.0 are not supported."
                except Exception:
                    pass
                self.fail(
                    msg.format(hint=hint), param, ctx)

            # Validate 'security' section keys
            security_config = safety_policy.get('security', {})
            security_keys = ['ignore-cvss-severity-below', 'ignore-cvss-unknown-severity', 'ignore-vulnerabilities',
                             'continue-on-vulnerability-error', 'ignore-unpinned-requirements']
            self.fail_if_unrecognized_keys(security_config.keys(), security_keys, param=param, ctx=ctx, msg=msg,
                                           context_hint='"security" -> ')

            # Validate 'ignore-cvss-severity-below' value
            ignore_cvss_security_below = security_config.get('ignore-cvss-severity-below', None)
            if ignore_cvss_security_below:
                limit = 0.0
                try:
                    limit = float(ignore_cvss_security_below)
                except ValueError as e:
                    self.fail(msg.format(hint="'ignore-cvss-severity-below' value needs to be an integer or float."))
                if limit < 0 or limit > 10:
                    self.fail(msg.format(hint="'ignore-cvss-severity-below' needs to be a value between 0 and 10"))

            # Validate 'continue-on-vulnerability-error' value
            continue_on_vulnerability_error = security_config.get('continue-on-vulnerability-error', None)
            self.fail_if_wrong_bool_value('continue-on-vulnerability-error', continue_on_vulnerability_error, msg)

            # Validate 'ignore-cvss-unknown-severity' value
            ignore_cvss_unknown_severity = security_config.get('ignore-cvss-unknown-severity', None)
            self.fail_if_wrong_bool_value('ignore-cvss-unknown-severity', ignore_cvss_unknown_severity, msg)

            # Validate 'ignore-vulnerabilities' section
            ignore_vulns = safety_policy.get('security', {}).get('ignore-vulnerabilities', {})
            if ignore_vulns:
                if not isinstance(ignore_vulns, dict):
                    self.fail(msg.format(hint="Vulnerability IDs under the 'ignore-vulnerabilities' key, need to "
                                              "follow the convention 'ID_NUMBER:', probably you are missing a colon."))

                normalized = {}

                for ignored_vuln_id, config in ignore_vulns.items():
                    ignored_vuln_config = config if config else {}

                    if not isinstance(ignored_vuln_config, dict):
                        self.fail(
                            msg.format(hint=f"Wrong configuration under the vulnerability with ID: {ignored_vuln_id}"))

                    context_msg = f'"security" -> "ignore-vulnerabilities" -> "{ignored_vuln_id}" -> '

                    self.fail_if_unrecognized_keys(ignored_vuln_config.keys(), ['reason', 'expires'], param=param,
                                                   ctx=ctx, msg=msg, context_hint=context_msg)

                    reason = ignored_vuln_config.get('reason', '')
                    reason = str(reason) if reason else None
                    expires = ignored_vuln_config.get('expires', '')
                    expires = str(expires) if expires else None

                    try:
                        if int(ignored_vuln_id) < 0:
                            raise ValueError('Negative Vulnerability ID')
                    except ValueError as e:
                        self.fail(msg.format(
                            hint=f"vulnerability id {ignored_vuln_id} under the 'ignore-vulnerabilities' root needs to "
                                 f"be a positive integer")
                        )

                    # Validate expires date
                    d = validate_expiration_date(expires)

                    if expires and not d:
                        self.fail(msg.format(hint=f"{context_msg}expires: \"{expires}\" isn't a valid format "
                                                  f"for the expires keyword, "
                                                  "valid options are: YYYY-MM-DD or "
                                                  "YYYY-MM-DD HH:MM:SS")
                                  )

                    normalized[str(ignored_vuln_id)] = {'reason': reason, 'expires': d}

                safety_policy['security']['ignore-vulnerabilities'] = normalized
                safety_policy['filename'] = filename
                safety_policy['raw'] = raw
            else:
                safety_policy['security']['ignore-vulnerabilities'] = {}

            # Validate 'fix' section keys
            fix_config = safety_policy.get('fix', {})
            self.fail_if_unrecognized_keys(fix_config.keys(), ['auto-security-updates-limit'], param=param, ctx=ctx, msg=msg, context_hint='"fix" -> ')
            auto_remediation_limit = fix_config.get('auto-security-updates-limit', None)

            if auto_remediation_limit:
                self.fail_if_unrecognized_keys(auto_remediation_limit, ['patch', 'minor', 'major'], param=param, ctx=ctx,
                                               msg=msg,
                                               context_hint='"auto-security-updates-limit" -> ')

            return safety_policy
        except BadParameter as expected_e:
            raise expected_e
        except Exception as e:
            # Handle file not found errors gracefully, don't fail in the default case
            if ctx and isinstance(e, OSError):
                default = ctx.get_parameter_source
                source = default("policy_file") if default("policy_file") else default("policy_file_path")
                if e.errno == 2 and source == click.core.ParameterSource.DEFAULT and value == '.safety-policy.yml':
                    return None

            problem = click.style("Policy file YAML is not valid.")
            hint = click.style("HINT: ", fg='yellow') + str(e)
            self.fail(f"{problem}\n{hint}", param, ctx)

    def shell_complete(
        self, ctx: click.Context, param: click.Parameter, incomplete: str
    ):
        """
        Return a special completion marker that tells the completion
        system to use the shell to provide file path completions.

        Args:
            ctx (click.Context): The click context.
            param (click.Parameter): The click parameter.
            incomplete (str): The value being completed. May be empty.

        Returns:
            List[click.shell_completion.CompletionItem]: The completion items.

        .. versionadded:: 8.0
        """
        from click.shell_completion import CompletionItem

        return [CompletionItem(incomplete, type="file")]


class SingletonMeta(type):
    """
    A metaclass for singleton classes.
    """

    _instances: Dict[type, Any] = {}

    _lock: Lock = Lock()

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SafetyContext(metaclass=SingletonMeta):
    """
    A singleton class to hold the Safety context.
    """
    packages = []
    key = False
    db_mirror = False
    cached = None
    ignore_vulns = None
    ignore_severity_rules = None
    proxy = None
    include_ignored = False
    telemetry = None
    files = None
    stdin = None
    is_env_scan = None
    command: Optional[str] = None
    subcommand: Optional[str] = None
    review = None
    params = {}
    safety_source = 'code'
    local_announcements = []
    scanned_full_path = []
    account = None



def sync_safety_context(f):
    """
    Decorator to sync the Safety context with the function arguments.
    """
    def new_func(*args, **kwargs):
        ctx = SafetyContext()

        legacy_key_added = False
        if "session" in kwargs:
            legacy_key_added = True
            kwargs["key"] = kwargs.get("session").api_key

        for attr in dir(ctx):
            if attr in kwargs:
                setattr(ctx, attr, kwargs.get(attr))

        if legacy_key_added:
            kwargs.pop("key")

        return f(*args, **kwargs)

    return new_func


@sync_safety_context
def get_packages_licenses(*, packages: Optional[List[Package]] = None, licenses_db: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
    """
    Get the licenses for the specified packages based on their version.

    Args:
        packages (Optional[List[Package]]): The list of packages.
        licenses_db (Optional[Dict[str, Any]]): The licenses database.

    Returns:
        List[Dict[str, Any]]: The list of packages and their licenses.
    """
    SafetyContext().command = 'license'

    if not packages:
        packages = []
    if not licenses_db:
        licenses_db = {}

    packages_licenses_db = licenses_db.get('packages', {})
    filtered_packages_licenses = []

    for pkg in packages:
        # Ignore recursive files not resolved
        if isinstance(pkg, RequirementFile):
            continue
        # normalize the package name
        pkg_name = canonicalize_name(pkg.name)
        # packages may have different licenses depending their version.
        pkg_licenses = packages_licenses_db.get(pkg_name, [])
        if not pkg.version:
            for req in pkg.requirements:
                if is_pinned_requirement(req.specifier):
                    pkg.version = next(iter(req.specifier)).version
                    break

            if not pkg.version:
                continue
        version_requested = parse_version(pkg.version)
        license_id = None
        license_name = None
        for pkg_version in pkg_licenses:
            license_start_version = parse_version(pkg_version['start_version'])
            # Stops and return the previous stored license when a new
            # license starts on a version above the requested one.
            if version_requested >= license_start_version:
                license_id = pkg_version['license_id']
            else:
                # We found the license for the version requested
                break

        if license_id:
            license_name = get_license_name_by_id(license_id, licenses_db)
        if not license_id or not license_name:
            license_name = "unknown"

        filtered_packages_licenses.append({
            "package": pkg_name,
            "version": pkg.version,
            "license": license_name
        })

    return filtered_packages_licenses


def get_requirements_content(files: List[click.File]) -> Dict[str, str]:
    """
    Get the content of the requirements files.

    Args:
        files (List[click.File]): The list of requirement files.

    Returns:
        Dict[str, str]: The content of the requirement files.

    Raises:
        InvalidProvidedReportError: If a file cannot be read.
    """
    requirements_files = {}

    for f in files:
        try:
            f.seek(0)
            requirements_files[f.name] = f.read()
            f.close()
        except Exception as e:
            raise InvalidProvidedReportError(message=f"Unable to read a requirement file scanned in the report. {e}")

    return requirements_files


def is_ignore_unpinned_mode(version: str) -> bool:
    """
    Check if unpinned mode is enabled based on the version.

    Args:
        version (str): The version string.

    Returns:
        bool: True if unpinned mode is enabled, False otherwise.
    """
    ignore = SafetyContext().params.get('ignore_unpinned_requirements')
    return (ignore is None or ignore) and not version


def get_remediations_count(remediations: Dict[str, Any]) -> int:
    """
    Get the count of remediations.

    Args:
        remediations (Dict[str, Any]): The remediations dictionary.

    Returns:
        int: The count of remediations.
    """
    return sum((len(rem.keys()) for pkg, rem in remediations.items()))


def get_hashes(dependency: Any) -> List[Dict[str, str]]:
    """
    Get the hashes for a dependency.

    Args:
        dependency (Any): The dependency.

    Returns:
        List[Dict[str, str]]: The list of hashes.
    """
    pattern = re.compile(HASH_REGEX_GROUPS)

    return [{'method': method, 'hash': hsh} for method, hsh in
               (pattern.match(d_hash).groups() for d_hash in dependency.hashes)]


def pluralize(word: str, count: int = 0) -> str:
    """
    Pluralize a word based on the count.

    Args:
        word (str): The word to pluralize.
        count (int): The count.

    Returns:
        str: The pluralized word.
    """
    if count == 1:
        return word

    default = {"was": "were", "this": "these", "has": "have"}

    if word in default:
        return default[word]

    if word.endswith("s") or word.endswith("x") or word.endswith("z") \
        or word.endswith("ch") or word.endswith("sh"):
        return word + "es"

    if word.endswith("y"):
        if word[-2] in "aeiou":
            return word + "s"
        else:
            return word[:-1] + "ies"

    return word + "s"


def initializate_config_dirs() -> None:
    """
    Initialize the configuration directories.
    """
    USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    try:
        SYSTEM_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        pass


import configparser
import logging
from pathlib import Path
import re
import time
from typing import Any, Dict, Generator, Optional, Set, Tuple, Union
from pydantic import ValidationError
import typer
from ..auth.utils import SafetyAuthSession
from ..errors import SafetyError
from .ecosystems.base import InspectableFile
from .ecosystems.target import InspectableFileContext
from .models import ScanExport, UnverifiedProjectModel

from safety_schemas.models import FileType, PolicyFileModel, PolicySource, \
    ConfigModel, Stage, ProjectModel, ScanType


LOG = logging.getLogger(__name__)

PROJECT_CONFIG = ".safety-project.ini"
PROJECT_CONFIG_SECTION = "project"
PROJECT_CONFIG_ID = "id"
PROJECT_CONFIG_URL = "url"
PROJECT_CONFIG_NAME = "name"


def download_policy(session: SafetyAuthSession, project_id: str, stage: Stage, branch: Optional[str]) -> Optional[PolicyFileModel]:
    """
    Downloads the policy file from the cloud for the given project and stage.

    Args:
        session (SafetyAuthSession): SafetyAuthSession object for authentication.
        project_id (str): The ID of the project.
        stage (Stage): The stage of the project.
        branch (Optional[str]): The branch of the project (optional).

    Returns:
        Optional[PolicyFileModel]: PolicyFileModel object if successful, otherwise None.
    """
    result = session.download_policy(project_id=project_id, stage=stage,
                                     branch=branch)

    if result and "uuid" in result and result["uuid"]:
        LOG.debug(f"Loading CLOUD policy file {result['uuid']} from cloud.")
        LOG.debug(result)
        uuid = result["uuid"]
        err = f'Unable to load the Safety Policy file ("{uuid}"), from cloud.'
        config = None

        try:
            yml_raw = result["settings"]
            # TODO: Move this to safety_schemas
            parse = "parse_obj"
            import importlib
            module_name = (
                "safety_schemas." "config.schemas." f"v3_0.main"
            )
            module = importlib.import_module(module_name)
            config_model = module.Config
            validated_policy_file = getattr(config_model, parse)(yml_raw)
            config = ConfigModel.from_v30(obj=validated_policy_file)
        except ValidationError as e:
            LOG.error(f"Failed to parse policy file {uuid}.", exc_info=True)
            raise SafetyError(f"{err}, details: {e}")
        except ValueError as e:
            LOG.error(f"Wrong YML file for policy file {uuid}.", exc_info=True)
            raise SafetyError(f"{err}, details: {e}")

        return PolicyFileModel(id=result["uuid"],
                                source=PolicySource.cloud,
                                location=None,
                                config=config)

    return None


def load_unverified_project_from_config(project_root: Path) -> UnverifiedProjectModel:
    """
    Loads an unverified project from the configuration file located at the project root.

    Args:
        project_root (Path): The root directory of the project.

    Returns:
        UnverifiedProjectModel: An instance of UnverifiedProjectModel.
    """
    config = configparser.ConfigParser()
    project_path = project_root / PROJECT_CONFIG
    config.read(project_path)
    id = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_ID, fallback=None)
    id = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_ID, fallback=None)
    url = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_URL, fallback=None)
    name = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_NAME, fallback=None)
    created = True
    if id:
        created = False

    return UnverifiedProjectModel(id=id, url_path=url,
                                  name=name, project_path=project_path,
                                  created=created)


def save_project_info(project: ProjectModel, project_path: Path) -> None:
    """
    Saves the project information to the configuration file.

    Args:
        project (ProjectModel): The ProjectModel object containing project information.
        project_path (Path): The path to the configuration file.
    """
    config = configparser.ConfigParser()
    config.read(project_path)

    if PROJECT_CONFIG_SECTION not in config.sections():
        config[PROJECT_CONFIG_SECTION] = {}

    config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_ID] = project.id
    if project.url_path:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_URL] = project.url_path
    if project.name:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_NAME] = project.name

    with open(project_path, 'w') as configfile:
        config.write(configfile)


def load_policy_file(path: Path) -> Optional[PolicyFileModel]:
    """
    Loads a policy file from the specified path.

    Args:
        path (Path): The path to the policy file.

    Returns:
        Optional[PolicyFileModel]: PolicyFileModel object if successful, otherwise None.
    """
    config = None

    if not path or not path.exists():
        return None

    err = f'Unable to load the Safety Policy file ("{path}"), this command ' \
        "only supports version 3.0"

    try:
        config = ConfigModel.parse_policy_file(raw_report=path)
    except ValidationError as e:
        LOG.error(f"Failed to parse policy file {path}.", exc_info=True)
        raise SafetyError(f"{err}, details: {e}")
    except ValueError as e:
        LOG.error(f"Wrong YML file for policy file {path}.", exc_info=True)
        raise SafetyError(f"{err}, details: {e}")

    return PolicyFileModel(id=str(path), source=PolicySource.local,
                           location=path, config=config)


def resolve_policy(local_policy: Optional[PolicyFileModel], cloud_policy: Optional[PolicyFileModel]) -> Optional[PolicyFileModel]:
    """
    Resolves the policy to be used, preferring cloud policy over local policy.

    Args:
        local_policy (Optional[PolicyFileModel]): The local policy file model (optional).
        cloud_policy (Optional[PolicyFileModel]): The cloud policy file model (optional).

    Returns:
        Optional[PolicyFileModel]: The resolved PolicyFileModel object.
    """
    policy = None

    if cloud_policy:
        policy = cloud_policy
    elif local_policy:
        policy = local_policy

    return policy


def save_report_as(scan_type: ScanType, export_type: ScanExport, at: Path, report: Any) -> None:
    """
    Saves the scan report to the specified location.

    Args:
        scan_type (ScanType): The type of scan.
        export_type (ScanExport): The type of export.
        at (Path): The path to save the report.
        report (Any): The report content.
    """
    tag = int(time.time())

    if at.is_dir():
        at = at / Path(
            f"{scan_type.value}-{export_type.get_default_file_name(tag=tag)}")

    with open(at, 'w+') as report_file:
        report_file.write(report)


def process_files(paths: Dict[str, Set[Path]], config: Optional[ConfigModel] = None) -> Generator[Tuple[Path, InspectableFile], None, None]:
    """
    Processes the files and yields each file path along with its inspectable file.

    Args:
        paths (Dict[str, Set[Path]]): A dictionary of file paths by file type.
        config (Optional[ConfigModel]): The configuration model (optional).

    Yields:
        Tuple[Path, InspectableFile]: A tuple of file path and inspectable file.
    """
    if not config:
        config = ConfigModel()

    for file_type_key, f_paths in paths.items():
        file_type = FileType(file_type_key)
        if not file_type or not file_type.ecosystem:
            continue
        for f_path in f_paths:
            with InspectableFileContext(f_path, file_type=file_type) as inspectable_file:
                if inspectable_file and inspectable_file.file_type:
                    inspectable_file.inspect(config=config)
                    inspectable_file.remediate()
                    yield f_path, inspectable_file



from datetime import datetime
import logging
import sys
from safety.auth.models import Auth

from safety.auth.utils import is_email_verified
from safety.console import main_console as console
from safety.constants import MSG_FINISH_REGISTRATION_TPL, MSG_VERIFICATION_HINT

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from typing import Optional

import click
from typer import Typer
import typer

from safety.auth.main import get_auth_info, get_authorization_data, get_token, clean_session
from safety.auth.server import process_browser_callback
from ..cli_util import get_command_for, pass_safety_cli_obj, SafetyCLISubGroup

from .constants import MSG_FAIL_LOGIN_AUTHED, MSG_FAIL_REGISTER_AUTHED, MSG_LOGOUT_DONE, MSG_LOGOUT_FAILED, MSG_NON_AUTHENTICATED
from safety.scan.constants import CLI_AUTH_COMMAND_HELP, CLI_AUTH_HEADLESS_HELP, DEFAULT_EPILOG, CLI_AUTH_LOGIN_HELP, CLI_AUTH_LOGOUT_HELP, CLI_AUTH_STATUS_HELP


from rich.padding import Padding

LOG = logging.getLogger(__name__)

auth_app = Typer(rich_markup_mode="rich")



CMD_LOGIN_NAME = "login"
CMD_REGISTER_NAME = "register"
CMD_STATUS_NAME = "status"
CMD_LOGOUT_NAME = "logout"
DEFAULT_CMD = CMD_LOGIN_NAME

@auth_app.callback(invoke_without_command=True,
                   cls=SafetyCLISubGroup,
                   help=CLI_AUTH_COMMAND_HELP,
                   epilog=DEFAULT_EPILOG,
                   context_settings={"allow_extra_args": True,
                                     "ignore_unknown_options": True})
@pass_safety_cli_obj
def auth(ctx: typer.Context) -> None:
    """
    Authenticate Safety CLI with your account.

    Args:
        ctx (typer.Context): The Typer context object.
    """
    LOG.info('auth started')

    # If no subcommand is invoked, forward to the default command
    if not ctx.invoked_subcommand:
        default_command = get_command_for(name=DEFAULT_CMD,
                                          typer_instance=auth_app)
        return ctx.forward(default_command)


def fail_if_authenticated(ctx: typer.Context, with_msg: str) -> None:
    """
    Exits the command if the user is already authenticated.

    Args:
        ctx (typer.Context): The Typer context object.
        with_msg (str): The message to display if authenticated.
    """
    info = get_auth_info(ctx)

    if info:
        console.print()
        email = f"[green]{ctx.obj.auth.email}[/green]"
        if not ctx.obj.auth.email_verified:
            email = f"{email} {render_email_note(ctx.obj.auth)}"

        console.print(with_msg.format(email=email))
        sys.exit(0)

def render_email_note(auth: Auth) -> str:
    """
    Renders a note indicating whether email verification is required.

    Args:
        auth (Auth): The Auth object.

    Returns:
        str: The rendered email note.
    """
    return "" if auth.email_verified else "[red](email verification required)[/red]"

def render_successful_login(auth: Auth,
                            organization: Optional[str] = None) -> None:
    """
    Renders a message indicating a successful login.

    Args:
        auth (Auth): The Auth object.
        organization (Optional[str]): The organization name.
    """
    DEFAULT = "--"
    name = auth.name if auth.name else DEFAULT
    email = auth.email if auth.email else DEFAULT
    email_note = render_email_note(auth)

    console.print("[bold][green]You're authenticated[/green][/bold]")
    if name and name != email:
        details = [f"[green][bold]Account:[/bold] {name}, {email}[/green] {email_note}"]
    else:
        details = [f"[green][bold]Account:[/bold] {email}[/green] {email_note}"]

    if organization:
        details.insert(0,
                        "[green][bold]Organization:[/bold] " \
                        f"{organization}[green]")

    for msg in details:
        console.print(Padding(msg, (0, 0, 0, 1)), emoji=True)


@auth_app.command(name=CMD_LOGIN_NAME, help=CLI_AUTH_LOGIN_HELP)
def login(
    ctx: typer.Context,
    headless: Annotated[
        Optional[bool],
        typer.Option(
            "--headless",
            help=CLI_AUTH_HEADLESS_HELP,
        )
    ] = None
) -> None:
    headless = headless is True 
    """
    Authenticate Safety CLI with your safetycli.com account using your default browser.

    Args:
        ctx (typer.Context): The Typer context object.
        headless (bool): Whether to run in headless mode.
    """
    LOG.info('login started')

    # Check if the user is already authenticated
    fail_if_authenticated(ctx, with_msg=MSG_FAIL_LOGIN_AUTHED)

    console.print()

    info = None

    brief_msg: str = "Redirecting your browser to log in; once authenticated, " \
            "return here to start using Safety"

    if ctx.obj.auth.org:
        console.print(f"Logging into [bold]{ctx.obj.auth.org.name}[/bold] " \
                      "organization.")

    if headless:
        brief_msg = "Running in headless mode. Please copy and open the following URL in a browser"

    # Get authorization data and generate the authorization URL
    uri, initial_state = get_authorization_data(client=ctx.obj.auth.client,
                                                code_verifier=ctx.obj.auth.code_verifier,
                                                organization=ctx.obj.auth.org, headless=headless)
    click.secho(brief_msg)
    click.echo()

    # Process the browser callback to complete the authentication
    info = process_browser_callback(uri, initial_state=initial_state, ctx=ctx, headless=headless)


    if info:
        if info.get("email", None):
            organization = None
            if ctx.obj.auth.org and ctx.obj.auth.org.name:
                organization = ctx.obj.auth.org.name
            ctx.obj.auth.refresh_from(info)
            if headless:
                console.print()

            render_successful_login(ctx.obj.auth, organization=organization)

            console.print()
            if ctx.obj.auth.org or ctx.obj.auth.email_verified:
                console.print(
                    "[tip]Tip[/tip]: now try [bold]`safety scan`[/bold] in your project’s root " \
                    "folder to run a project scan or [bold]`safety -–help`[/bold] to learn more.")
            else:
                console.print(MSG_FINISH_REGISTRATION_TPL.format(email=ctx.obj.auth.email))
                console.print()
                console.print(MSG_VERIFICATION_HINT)
        else:
            click.secho("Safety is now authenticated but your email is missing.")
    else:
        msg = ":stop_sign: [red]"
        if ctx.obj.auth.org:
            msg += f"Error logging into {ctx.obj.auth.org.name} organization " \
                f"with auth ID: {ctx.obj.auth.org.id}."
        else:
            msg += "Error logging into Safety."

        msg += " Please try again, or use [bold]`safety auth -–help`[/bold] " \
            "for more information[/red]"

        console.print(msg, emoji=True)

@auth_app.command(name=CMD_LOGOUT_NAME, help=CLI_AUTH_LOGOUT_HELP)
def logout(ctx: typer.Context) -> None:
    """
    Log out of your current session.

    Args:
        ctx (typer.Context): The Typer context object.
    """
    LOG.info('logout started')

    id_token = get_token('id_token')

    msg = MSG_NON_AUTHENTICATED

    if id_token:
         # Clean the session if an ID token is found
        if clean_session(ctx.obj.auth.client):
            msg = MSG_LOGOUT_DONE
        else:
            msg = MSG_LOGOUT_FAILED

    console.print(msg)


@auth_app.command(name=CMD_STATUS_NAME, help=CLI_AUTH_STATUS_HELP)
@click.option("--ensure-auth/--no-ensure-auth", default=False,
              help="This will keep running the command until an" \
                "authentication is made.")
@click.option("--login-timeout", "-w", type=int, default=600,
              help="Max time allowed to wait for an authentication.")
def status(ctx: typer.Context, ensure_auth: bool = False,
           login_timeout: int = 600)  -> None:
    """
    Display Safety CLI's current authentication status.

    Args:
        ctx (typer.Context): The Typer context object.
        ensure_auth (bool): Whether to keep running until authentication is made.
        login_timeout (int): Max time allowed to wait for authentication.
    """
    LOG.info('status started')
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    from safety.util import get_safety_version
    safety_version = get_safety_version()
    console.print(f"[{current_time}]: Safety {safety_version}")

    info = get_auth_info(ctx)

    if ensure_auth:
        console.print("running: safety auth status --ensure-auth")
        console.print()

    if info:
        verified = is_email_verified(info)
        email_status = " [red](email not verified)[/red]" if not verified else ""

        console.print(f'[green]Authenticated as {info["email"]}[/green]{email_status}')
    elif ensure_auth:
        console.print('Safety is not authenticated. Launching default browser to log in')
        console.print()
        uri, initial_state = get_authorization_data(client=ctx.obj.auth.client,
                                                    code_verifier=ctx.obj.auth.code_verifier,
                                                    organization=ctx.obj.auth.org, ensure_auth=ensure_auth)

        # Process the browser callback to complete the authentication
        info = process_browser_callback(uri, initial_state=initial_state,
                                          timeout=login_timeout, ctx=ctx)

        if not info:
            console.print(f'[red]Timeout error ({login_timeout} seconds): not successfully authenticated without the timeout period.[/red]')
            sys.exit(1)

        organization = None
        if ctx.obj.auth.org and ctx.obj.auth.org.name:
            organization = ctx.obj.auth.org.name

        render_successful_login(ctx.obj.auth, organization=organization)
        console.print()

    else:
        console.print(MSG_NON_AUTHENTICATED)


@auth_app.command(name=CMD_REGISTER_NAME)
def register(ctx: typer.Context) -> None:
    """
    Create a new user account for the safetycli.com service.

    Args:
        ctx (typer.Context): The Typer context object.
    """
    LOG.info('register started')

    # Check if the user is already authenticated
    fail_if_authenticated(ctx, with_msg=MSG_FAIL_REGISTER_AUTHED)

    # Get authorization data and generate the registration URL
    uri, initial_state = get_authorization_data(client=ctx.obj.auth.client,
                                                code_verifier=ctx.obj.auth.code_verifier,
                                                sign_up=True)

    console.print("Redirecting your browser to register for a free account. Once registered, return here to start using Safety.")
    console.print()

    # Process the browser callback to complete the registration
    info = process_browser_callback(uri,
                                    initial_state=initial_state, ctx=ctx)

    if info:
        console.print(f'[green]Successfully registered {info.get("email")}[/green]')
        console.print()
    else:
        console.print('[red]Unable to register in this time, try again.[/red]')




import os
import unittest

from unittest.mock import patch, Mock
from click.testing import CliRunner
from safety.cli import cli

from safety.scan.command import scan
from safety.scan.command import scan_project_app

class TestScanCommand(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner(mix_stderr=False)
        self.dirname = os.path.dirname(__file__)

    def test_scan(self):
        result = self.runner.invoke(cli, ["--stage", "cicd", "scan", "--target", self.dirname, "--output", "json"])
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(cli, ["--stage", "production", "scan", "--target", self.dirname, "--output", "json"])
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(cli, ["--stage", "cicd", "scan", "--target", self.dirname, "--output", "screen"])
        self.assertEqual(result.exit_code, 1)



import os
import unittest
from unittest.mock import MagicMock, Mock, call, patch
from pathlib import Path
import datetime

from safety.scan.render import print_announcements, print_summary, render_header
from safety_schemas.models import ProjectModel, IgnoreCodes, PolicySource

class TestRender(unittest.TestCase):

    def setUp(self):
        self.console = MagicMock()
        self.project = ProjectModel(id='test-project')
        self.project.policy = MagicMock()
        self.project.policy.source = PolicySource.cloud

    @patch('safety.scan.render.get_safety_version')
    def test_render_header(self, mock_get_safety_version):
        mock_get_safety_version.return_value = '3.0.0'

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime(2025, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

        test_cases = [
            ([Path('/target1'), Path('/target2')], False, 'Safety 3.0.0 scanning {0}, {1}\n2025-01-01 00:00:00 UTC'),
            ([Path('/target1')], False, 'Safety 3.0.0 scanning {0}\n2025-01-01 00:00:00 UTC'),
            ([], False, 'Safety 3.0.0 scanning \n2025-01-01 00:00:00 UTC'),
            ([Path('/target1'), Path('/target2')], True, 'Safety 3.0.0 running system scan\n2025-01-01 00:00:00 UTC'),
        ]

        for targets, is_system_scan, expected_result in test_cases:
            with patch('datetime.datetime', new=datetime_mock):
                # Normalize paths to POSIX format (use forward slashes)
                posix_targets = [t.as_posix() for t in targets]
                result = str(render_header(targets, is_system_scan))

                # On Windows, convert backslashes to forward slashes in the result for comparison
                if os.name == 'nt':
                    result = result.replace('\\', '/')

                expected = expected_result.format(*posix_targets)
                self.assertEqual(result, expected)

    @patch('safety.scan.render.safety.get_announcements')
    @patch('safety.scan.render.get_basic_announcements')
    @patch('safety.scan.render.Console')
    def test_print_announcements(self, mock_console, mock_get_basic_announcements, mock_get_announcements):
        
        mock_get_announcements.return_value = [
            {'type': 'info', 'message': 'Info message'},
            {'type': 'warning', 'message': 'Warning message'},
            {'type': 'error', 'message': 'Error message'},
        ]
        
        mock_get_basic_announcements.return_value = mock_get_announcements.return_value
        console = mock_console.return_value
        
        ctx = MagicMock()
        ctx.obj.auth.client = MagicMock()
        ctx.obj.config.telemetry_enabled = False
        ctx.obj.telemetry = MagicMock()
        
        print_announcements(console, ctx)

        # Include empty calls and correct sequence
        console.print.assert_has_calls([
            call(),
            call("[bold]Safety Announcements:[/bold]"),
            call(),
            call("[default]* Info message[/default]"),
            call("[yellow]* Warning message[/yellow]"),
            call("[red]* Error message[/red]")
        ])


    @patch('safety.scan.render.render_to_console')
    def test_print_summary(self, mock_render_to_console):
        ignored_vulns_data = [
            MagicMock(ignored_code=IgnoreCodes.manual.value, vulnerability_id='v1', package_name='p1'),
            MagicMock(ignored_code=IgnoreCodes.cvss_severity.value, vulnerability_id='v2', package_name='p2'),
            MagicMock(ignored_code=IgnoreCodes.unpinned_specification.value, vulnerability_id='v3', package_name='p3'),
            MagicMock(ignored_code=IgnoreCodes.environment_dependency.value, vulnerability_id='v4', package_name='p4'),
        ]

        print_summary(
            self.console,
            total_issues_with_duplicates=0,
            total_ignored_issues=0,
            project=self.project,
            dependencies_count=5,
            fixes_count=0,
            resolved_vulns_per_fix=0,
            ignored_vulns_data=ignored_vulns_data
        )

        self.console.print.assert_has_calls([
            call('Tested [number]5[/number] dependencies for security issues using policy fetched from Safety Platform'),
            call('0 security issues found, 0 fixes suggested.'),
            call('[number]0[/number] fixes suggested, resolving [number]0[/number] vulnerabilities.')
        ])

        print_summary(
            self.console,
            total_issues_with_duplicates=0,
            total_ignored_issues=0,
            project=self.project,
            dependencies_count=5,
            fixes_count=0,
            resolved_vulns_per_fix=0
        )

        self.console.print.assert_has_calls([
            call('Tested [number]5[/number] dependencies for security issues using policy fetched from Safety Platform'),
            call('0 security issues found, 0 fixes suggested.'),
            call('[number]0[/number] fixes suggested, resolving [number]0[/number] vulnerabilities.')
        ])

if __name__ == '__main__':
    unittest.main()



import os
import sys
import unittest
from io import StringIO
from unittest.mock import patch, Mock

import click as click

from safety import util
from safety.models import SafetyRequirement
from safety.util import read_requirements, get_processed_options, SafetyPolicyFile, transform_ignore


class ReadRequirementsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dirname = os.path.dirname(__file__)

    def test_unpinned_vcs_requirement(self):
        """
        https://github.com/pyupio/safety/issues/72
        """
        # this shouldn't raise an error
        content = StringIO("-e git+https://github.com/jdunck/python-unicodecsv#egg=unicodecsv")
        result = list(read_requirements(content))
        self.assertEqual(len(result), 0)

    def test_recursive_requirement(self):
        """
        https://github.com/pyupio/safety/issues/132
        """
        # this should find 1 packages with two requirements found
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_1.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 1)
        found_pkg = result[0]
        self.assertEqual(found_pkg.name, 'insecure-package')
        self.assertEqual(found_pkg.version, None)
        expected = [SafetyRequirement('insecure-package==0.1.0'), SafetyRequirement('insecure-package==0.1.1')]
        self.assertListEqual(expected, found_pkg.requirements)

    def test_recursive_requirement_pinned_after_unpinned(self):
        # this should find 4 packages, unpinned aren't ignored.
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_pinned_and_unpinned.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 4)

    @patch("safety.util.get_flags_from_context")
    @patch.object(sys, 'argv', ['safety/__main__.py', 'check', '--key=my-key', '-i', '3232', '-i', '3231', '--ignore',
                                '1212', '--json'])
    def test_log_used_options_with_argv(self, get_flags_from_context):
        get_flags_from_context.return_value = {'--key': 'key', '--db': 'db', '--json': 'json', '--no-json': 'json',
                                               '--full-report': 'full_report', '--short-report': 'full_report',
                                               '--bare': 'bare', '--not-bare': 'bare', '--cache': 'cache',
                                               '--no-cache': 'cache', '--stdin': 'stdin', '--no-stdin': 'stdin',
                                               '--file': 'files', '-r': 'files', '--ignore': 'ignore', '-i': 'ignore',
                                               '--output': 'output', '-o': 'output', '--proxy-host': 'proxyhost',
                                               '-ph': 'proxyhost', '--proxy-port': 'proxyport', '-pp': 'proxyport',
                                               '--proxy-protocol': 'proxyprotocol', '-pr': 'proxyprotocol'}
        used_options = util.get_used_options()

        self.assertEqual(used_options, {'ignore': {'-i': 2, '--ignore': 1}, 'json': {'--json': 1}, 'key': {'--key': 1}})

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)))
    def test_cli_ignore_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_with_ignores.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        cli_ignores = {'1234': {'reason': '', 'expires': None}}
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = get_processed_options(
            policy_file=policy_file, ignore=cli_ignores,
            ignore_severity_rules=None, exit_code=True)

        self.assertEqual(ignore, cli_ignores)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.COMMANDLINE)))
    def test_cli_continue_on_error_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = get_processed_options(
            policy_file=policy_file, ignore={},
            ignore_severity_rules=None, exit_code=True)

        self.assertEqual(exit_code, True)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.COMMANDLINE)))
    def test_cli_exit_code_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the exit_code argument by commandline
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = get_processed_options(
            policy_file=policy_file, ignore={},
            ignore_severity_rules=None, exit_code=True)

        security_pf = policy_file.get('security', {})
        severity_rules = {'ignore-cvss-severity-below': security_pf.get('ignore-cvss-severity-below', 0.0),
                          'ignore-cvss-unknown-severity': security_pf.get('ignore-cvss-unknown-severity', False)}

        self.assertEqual(ignore, security_pf.get('ignore-vulnerabilities', None))
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertEqual(exit_code, True)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)))
    def test_cli_ignore_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the ignores argument by commandline
        cli_ignores = {'1234': {'reason': '', 'expires': None}}
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = get_processed_options(
            policy_file=policy_file, ignore=cli_ignores,
            ignore_severity_rules=None, exit_code=True)

        security_pf = policy_file.get('security', {})
        severity_rules = {'ignore-cvss-severity-below': security_pf.get('ignore-cvss-severity-below', 0.0),
                          'ignore-cvss-unknown-severity': security_pf.get('ignore-cvss-unknown-severity', False)}

        self.assertEqual(ignore, cli_ignores)
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertIsNotNone(security_pf.get('continue-on-vulnerability-error', None),
                             msg='This test requires a yml with a continue-on-vulnerability-error value')
        EXPECTED = not security_pf.get('continue-on-vulnerability-error')
        self.assertEqual(exit_code, EXPECTED)

    def test_transform_ignore(self):
        ignored_transformed = {'123': {'expires': None, 'reason': ''}, '456': {'expires': None, 'reason': ''}}
        self.assertEqual(transform_ignore(None, None, value=("123", "456")), ignored_transformed)
        self.assertEqual(transform_ignore(None, None, value=("123,456",)), ignored_transformed)

    def test_transform_ignore_mixed_arguments(self):
        # mix old and new way of providing --ignore
        ignored_transformed = {
            '123': {'expires': None, 'reason': ''},
            '456': {'expires': None, 'reason': ''},
            '789': {'expires': None, 'reason': ''}
        }
        self.assertEqual(transform_ignore(None, None, value=("123,456", "789")), ignored_transformed)




import os
from pathlib import Path
from typing import Optional, Tuple
import typer
from safety.scan.main import save_project_info
from safety.scan.models import ScanExport, ScanOutput, UnverifiedProjectModel
from safety.scan.render import print_wait_project_verification, prompt_project_id, prompt_link_project

from safety_schemas.models import AuthenticationType, ProjectModel, Stage
from safety.auth.utils import SafetyAuthSession


MISSING_SPDX_EXTENSION_MSG = "spdx extra is not installed, please install it with: pip install safety[spdx]"


def raise_if_not_spdx_extension_installed() -> None:
    """
    Raises an error if the spdx extension is not installed.
    """
    try:
        import spdx_tools.spdx
    except Exception as e:
        raise typer.BadParameter(MISSING_SPDX_EXTENSION_MSG)


def save_as_callback(save_as: Optional[Tuple[ScanExport, Path]]) -> Tuple[Optional[str], Optional[Path]]:
    """
    Callback function to handle save_as parameter and validate if spdx extension is installed.

    Args:
        save_as (Optional[Tuple[ScanExport, Path]]): The export type and path.

    Returns:
        Tuple[Optional[str], Optional[Path]]: The validated export type and path.
    """
    export_type, export_path = save_as if save_as else (None, None)

    if ScanExport.is_format(export_type, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()

    return (export_type.value, export_path) if export_type and export_path else (export_type, export_path)

def output_callback(output: ScanOutput) -> str:
    """
    Callback function to handle output parameter and validate if spdx extension is installed.

    Args:
        output (ScanOutput): The output format.

    Returns:
        str: The validated output format.
    """
    if ScanOutput.is_format(output, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()

    return output.value


def save_verified_project(ctx: typer.Context, slug: str, name: Optional[str], project_path: Path, url_path: Optional[str]):
    """
    Save the verified project information to the context and project info file.

    Args:
        ctx (typer.Context): The context of the Typer command.
        slug (str): The project slug.
        name (Optional[str]): The project name.
        project_path (Path): The project path.
        url_path (Optional[str]): The project URL path.
    """
    ctx.obj.project = ProjectModel(
        id=slug,
        name=name,
        project_path=project_path,
        url_path=url_path
    )
    if ctx.obj.auth.stage is Stage.development:
        save_project_info(project=ctx.obj.project,
                          project_path=project_path)


def check_project(console, ctx: typer.Context, session: SafetyAuthSession,
                  unverified_project: UnverifiedProjectModel, stage: Stage,
                  git_origin: Optional[str], ask_project_id: bool = False) -> dict:
    """
    Check the project against the session and stage, verifying the project if necessary.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        stage (Stage): The current stage.
        git_origin (Optional[str]): The Git origin URL.
        ask_project_id (bool): Whether to prompt for the project ID.

    Returns:
        dict: The result of the project check.
    """
    stage = ctx.obj.auth.stage
    source = ctx.obj.telemetry.safety_source if ctx.obj.telemetry else None
    data = {"scan_stage": stage, "safety_source": source}

    PRJ_SLUG_KEY = "project_slug"
    PRJ_SLUG_SOURCE_KEY = "project_slug_source"
    PRJ_GIT_ORIGIN_KEY = "git_origin"

    if git_origin:
        data[PRJ_GIT_ORIGIN_KEY] = git_origin

    if unverified_project.id:
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = ".safety-project.ini"
    elif not git_origin or ask_project_id:
        # Set a project id for this scan (no spaces). If empty Safety will use: pyupio:
        parent_root_name = None
        if unverified_project.project_path.parent.name:
            parent_root_name = unverified_project.project_path.parent.name

        unverified_project.id = prompt_project_id(console, stage, parent_root_name)
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = "user"

    status = print_wait_project_verification(console, data[PRJ_SLUG_KEY] if data.get(PRJ_SLUG_KEY, None) else "-",
                                    (session.check_project, data), on_error_delay=1)

    return status


def verify_project(console, ctx: typer.Context, session: SafetyAuthSession,
                   unverified_project: UnverifiedProjectModel, stage: Stage,
                   git_origin: Optional[str]):
    """
    Verify the project, linking it if necessary and saving the verified project information.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        stage (Stage): The current stage.
        git_origin (Optional[str]): The Git origin URL.
    """

    verified_prj = False

    link_prj = True

    while not verified_prj:
        result = check_project(console, ctx, session, unverified_project, stage, git_origin, ask_project_id=not link_prj)

        unverified_slug = result.get("slug")

        project = result.get("project", None)
        user_confirm = result.get("user_confirm", False)

        if user_confirm:
            if project and link_prj:
                prj_name = project.get("name", None)
                prj_admin_email = project.get("admin", None)

                link_prj = prompt_link_project(prj_name=prj_name,
                                    prj_admin_email=prj_admin_email,
                                    console=console)

                if not link_prj:
                    continue

        verified_prj = print_wait_project_verification(
            console, unverified_slug, (session.project,
                                       {"project_id": unverified_slug}),
                                       on_error_delay=1)

        if verified_prj and isinstance(verified_prj, dict) and verified_prj.get("slug", None):
            save_verified_project(ctx, verified_prj["slug"], verified_prj.get("name", None),
                                  unverified_project.project_path, verified_prj.get("url", None))
        else:
            verified_prj = False



from functools import lru_cache
import json
import logging
import sys
from typing import Any, Optional, Dict, Callable, Tuple
from authlib.integrations.requests_client import OAuth2Session
from authlib.integrations.base_client.errors import OAuthError
import requests
from requests.adapters import HTTPAdapter

from safety.auth.constants import AUTH_SERVER_URL, CLAIM_EMAIL_VERIFIED_API, \
    CLAIM_EMAIL_VERIFIED_AUTH_SERVER
from safety.auth.main import get_auth_info, get_token_data
from safety.constants import PLATFORM_API_CHECK_UPDATES_ENDPOINT, PLATFORM_API_INITIALIZE_SCAN_ENDPOINT, PLATFORM_API_POLICY_ENDPOINT, \
    PLATFORM_API_PROJECT_CHECK_ENDPOINT, PLATFORM_API_PROJECT_ENDPOINT, PLATFORM_API_PROJECT_SCAN_REQUEST_ENDPOINT, \
        PLATFORM_API_PROJECT_UPLOAD_SCAN_ENDPOINT, REQUEST_TIMEOUT
from safety.scan.util import AuthenticationType

from safety.util import SafetyContext, output_exception
from safety_schemas.models import STAGE_ID_MAPPING, Stage
from safety.errors import InvalidCredentialError, NetworkConnectionError, \
    RequestTimeoutError, ServerError, TooManyRequestsError, SafetyError

LOG = logging.getLogger(__name__)


def get_keys(client_session: OAuth2Session, openid_config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """
    Retrieve the keys from the OpenID configuration.

    Args:
        client_session (OAuth2Session): The OAuth2 session.
        openid_config (Dict[str, Any]): The OpenID configuration.

    Returns:
        Optional[Dict[str, Any]]: The keys, if available.
    """
    if "jwks_uri" in openid_config:
        return client_session.get(url=openid_config["jwks_uri"], bearer=False).json()
    return None


def is_email_verified(info: Dict[str, Any]) -> Optional[bool]:
    """
    Check if the email is verified.

    Args:
        info (Dict[str, Any]): The user information.

    Returns:
        bool: True if the email is verified, False otherwise.
    """
    return info.get(CLAIM_EMAIL_VERIFIED_API) or info.get(CLAIM_EMAIL_VERIFIED_AUTH_SERVER)


def parse_response(func: Callable) -> Callable:
    """
    Decorator to parse the response from an HTTP request.

    Args:
        func (Callable): The function to wrap.

    Returns:
        Callable: The wrapped function.
    """
    def wrapper(*args, **kwargs):
        try:
            r = func(*args, **kwargs)
        except OAuthError as e:
            LOG.exception('OAuth failed: %s', e)
            raise InvalidCredentialError(message="Your token authentication expired, try login again.")
        except requests.exceptions.ConnectionError:
            raise NetworkConnectionError()
        except requests.exceptions.Timeout:
            raise RequestTimeoutError()
        except requests.exceptions.RequestException as e:
            raise e

        if r.status_code == 403:
            raise InvalidCredentialError(credential="Failed authentication.",
                                               reason=r.text)

        if r.status_code == 429:
            raise TooManyRequestsError(reason=r.text)

        if r.status_code >= 400 and r.status_code < 500:
            error_code = None
            try:
                data = r.json()
                reason = data.get('detail', "Unable to find reason.")
                error_code = data.get("error_code", None)
            except Exception as e:
                reason = r.reason

            raise SafetyError(message=reason, error_code=error_code)

        if r.status_code >= 500:
            raise ServerError(reason=f"{r.reason} - {r.text}")

        data = None

        try:
            data = r.json()
        except json.JSONDecodeError as e:
            raise SafetyError(message=f"Bad JSON response: {e}")

        return data

    return wrapper

class SafetyAuthSession(OAuth2Session):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the SafetyAuthSession.

        Args:
            *args (Any): Positional arguments for the parent class.
            **kwargs (Any): Keyword arguments for the parent class.
        """
        super().__init__(*args, **kwargs)
        self.proxy_required: bool = False
        self.proxy_timeout: Optional[int] = None
        self.api_key = None

    def get_credential(self) -> Optional[str]:
        """
        Get the current authentication credential.

        Returns:
            Optional[str]: The API key, token, or None.
        """
        if self.api_key:
            return self.api_key

        if self.token:
            return SafetyContext().account

        return None

    def is_using_auth_credentials(self) -> bool:
        """
        Check if the session is using authentication credentials.

        This does NOT check if the client is authenticated.

        Returns:
            bool: True if using authentication credentials, False otherwise.
        """
        return self.get_authentication_type() != AuthenticationType.none

    def get_authentication_type(self) -> AuthenticationType:
        """
        Get the type of authentication being used.

        Returns:
            AuthenticationType: The type of authentication.
        """
        if self.api_key:
            return AuthenticationType.api_key

        if self.token:
            return AuthenticationType.token

        return AuthenticationType.none

    def request(self, method: str, url: str, withhold_token: bool = False, auth: Optional[Tuple] = None, bearer: bool = True, **kwargs: Any) -> requests.Response:
        """
        Make an HTTP request with the appropriate authentication.

        Use the right auth parameter for Safety supported auth types.

        Args:
            method (str): The HTTP method.
            url (str): The URL to request.
            withhold_token (bool): Whether to withhold the token.
            auth (Optional[Tuple]): The authentication tuple.
            bearer (bool): Whether to use bearer authentication.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            requests.Response: The HTTP response.

        Raises:
            Exception: If the request fails.
        """
        # By default use the token_auth
        TIMEOUT_KEYWARD = "timeout"
        func_timeout = kwargs[TIMEOUT_KEYWARD] if TIMEOUT_KEYWARD in kwargs else REQUEST_TIMEOUT

        if self.api_key:
            key_header = {"X-Api-Key": self.api_key}
            if not "headers" in kwargs:
                kwargs["headers"] = key_header
            else:
                kwargs["headers"]["X-Api-Key"] = self.api_key

        if not self.token or not bearer:
            # Fallback to no token auth
            auth = ()


        # Override proxies
        if self.proxies:
            kwargs['proxies'] = self.proxies

            if self.proxy_timeout:
                kwargs['timeout'] = int(self.proxy_timeout) / 1000

        if ("proxies" not in kwargs or not self.proxies) and self.proxy_required:
            output_exception("Proxy connection is required but there is not a proxy setup.", exit_code_output=True)

        request_func = super(SafetyAuthSession, self).request
        params = {
            'method': method,
            'url': url,
            'withhold_token': withhold_token,
            'auth': auth,
        }
        params.update(kwargs)

        try:
            return request_func(**params)
        except Exception as e:
            LOG.debug('Request failed: %s', e)

            if self.proxy_required:
                output_exception(f"Proxy is required but the connection failed because: {e}", exit_code_output=True)

            if ("proxies" in kwargs or self.proxies):
                params["proxies"] = {}
                params['timeout'] = func_timeout
                self.proxies = {}
                message = "The proxy configuration failed to function and was disregarded."
                LOG.debug(message)
                if message not in [a['message'] for a in SafetyContext.local_announcements]:
                    SafetyContext.local_announcements.append({'message': message, 'type': 'warning', 'local': True})

                return request_func(**params)

            raise e

    @parse_response
    def fetch_user_info(self) -> Any:
        """
        Fetch user information from the authorization server.

        Returns:
            Any: The user information.
        """
        USER_INFO_ENDPOINT = f"{AUTH_SERVER_URL}/userinfo"

        r = self.get(
            url=USER_INFO_ENDPOINT
        )

        return r

    @parse_response
    def check_project(self, scan_stage: str, safety_source: str,
                      project_slug: Optional[str] = None, git_origin: Optional[str] = None,
                      project_slug_source: Optional[str] = None) -> Any:
        """
        Check project information.

        Args:
            scan_stage (str): The scan stage.
            safety_source (str): The safety source.
            project_slug (Optional[str]): The project slug.
            git_origin (Optional[str]): The git origin.
            project_slug_source (Optional[str]): The project slug source.

        Returns:
            Any: The project information.
        """

        data = {"scan_stage": scan_stage, "safety_source": safety_source,
                "project_slug": project_slug,
                "project_slug_source": project_slug_source,
                "git_origin": git_origin}

        r = self.post(
                url=PLATFORM_API_PROJECT_CHECK_ENDPOINT,
                json=data
            )

        return r

    @parse_response
    def project(self, project_id: str) -> Any:
        """
        Get project information.

        Args:
            project_id (str): The project ID.

        Returns:
            Any: The project information.
        """
        data = {"project": project_id}

        return self.get(
            url=PLATFORM_API_PROJECT_ENDPOINT,
            params=data
        )

    @parse_response
    def download_policy(self, project_id: Optional[str], stage: Stage, branch: Optional[str]) -> Any:
        """
        Download the project policy.

        Args:
            project_id (Optional[str]): The project ID.
            stage (Stage): The stage.
            branch (Optional[str]): The branch.

        Returns:
            Any: The policy data.
        """
        data = {"project": project_id, "stage": STAGE_ID_MAPPING[stage], "branch": branch}

        return self.get(
            url=PLATFORM_API_POLICY_ENDPOINT,
            params=data
        )


    @parse_response
    def project_scan_request(self, project_id: str) -> Any:
        """
        Request a project scan.

        Args:
            project_id (str): The project ID.

        Returns:
            Any: The scan request result.
        """
        data = {"project_id": project_id}

        return self.post(
            url=PLATFORM_API_PROJECT_SCAN_REQUEST_ENDPOINT,
            json=data
        )


    @parse_response
    def upload_report(self, json_report: str) -> Any:
        """
        Upload a scan report.

        Args:
            json_report (str): The JSON report.

        Returns:
            Any: The upload result.
        """

        headers = {
            "Content-Type": "application/json"
        }

        return self.post(
            url=PLATFORM_API_PROJECT_UPLOAD_SCAN_ENDPOINT,
            data=json_report,
            headers=headers
        )


    @parse_response
    def check_updates(self, version: int, safety_version: Optional[str] = None, python_version: Optional[str] = None, os_type: Optional[str] = None, os_release: Optional[str] = None, os_description: Optional[str] = None) -> Any:
        """
        Check for updates.

        Args:
            version (int): The version.
            safety_version (Optional[str]): The Safety version.
            python_version (Optional[str]): The Python version.
            os_type (Optional[str]): The OS type.
            os_release (Optional[str]): The OS release.
            os_description (Optional[str]): The OS description.

        Returns:
            Any: The update check result.
        """
        data = {"version": version,
                "safety_version": safety_version,
                "python_version": python_version,
                "os_type": os_type,
                "os_release": os_release,
                "os_description": os_description}

        return self.get(
            url=PLATFORM_API_CHECK_UPDATES_ENDPOINT,
            params=data
        )


    @parse_response
    def initialize_scan(self) -> Any:
        """
        Initialize a scan.

        Returns:
            Any: The initialization result.
        """
        try:
            response = self.get(url=PLATFORM_API_INITIALIZE_SCAN_ENDPOINT, timeout=5)
            return response
        except requests.exceptions.Timeout:
            LOG.error("Auth request to initialize scan timed out after 5 seconds.")
        except Exception as e:
            LOG.exception("Exception trying to auth initialize scan", exc_info=True)
        return None

class S3PresignedAdapter(HTTPAdapter):
    def send(self, request: requests.PreparedRequest, **kwargs: Any) -> requests.Response:
        """
        Send a request, removing the Authorization header.

        Args:
            request (requests.PreparedRequest): The prepared request.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            requests.Response: The response.
        """
        request.headers.pop("Authorization", None)
        return super().send(request, **kwargs)
    
    
from functools import lru_cache

@lru_cache(maxsize=1)
def is_jupyter_notebook() -> bool:
    """
    Detects if the code is running in a Jupyter notebook environment, including
    various cloud-hosted Jupyter notebooks.

    Returns:
        bool: True if the environment is identified as a Jupyter notebook (or 
              equivalent cloud-based environment), False otherwise.

    Supported environments:
    - Google Colab
    - Amazon SageMaker
    - Azure Notebooks
    - Kaggle Notebooks
    - Databricks Notebooks
    - Datalore by JetBrains
    - Paperspace Gradient Notebooks
    - Classic Jupyter Notebook and JupyterLab

    Detection is done by attempting to import environment-specific packages:
    - Google Colab: `google.colab`
    - Amazon SageMaker: `sagemaker`
    - Azure Notebooks: `azureml`
    - Kaggle Notebooks: `kaggle`
    - Databricks Notebooks: `dbutils`
    - Datalore: `datalore`
    - Paperspace Gradient: `gradient`
    - Classic Jupyter: Checks if 'IPKernelApp' is in IPython config.

    Example:
        >>> is_jupyter_notebook()
        True
    """
    try:
        # Detect Google Colab
        import google.colab
        return True
    except ImportError:
        pass

    try:
        # Detect Amazon SageMaker
        import sagemaker
        return True
    except ImportError:
        pass

    try:
        # Detect Azure Notebooks
        import azureml
        return True
    except ImportError:
        pass

    try:
        # Detect Kaggle Notebooks
        import kaggle
        return True
    except ImportError:
        pass

    try:
        # Detect Databricks
        import dbutils
        return True
    except ImportError:
        pass

    try:
        # Detect Datalore
        import datalore
        return True
    except ImportError:
        pass

    try:
        # Detect Paperspace Gradient Notebooks
        import gradient
        return True
    except ImportError:
        pass

    try:
        # Detect classic Jupyter Notebook, JupyterLab, and other IPython kernel-based environments
        from IPython import get_ipython
        if 'IPKernelApp' in get_ipython().config:
            return True
    except:
        pass

    return False
